/************************************************************

     selectinfo.c  -- Main program for the selectinfo
      selection display program for X.


Copyright 1993, Hewlett-Packard Company.

                        All Rights Reserved

Permission to use, copy, modify, and distribute this software and to
use, copy, and distribute its documentation for any purpose and without
fee is hereby granted, provided that the above copyright notice appear
in all copies and that both that copyright notice and this permission
notice appear in supporting documentation, and that the name of
Hewlett-Packard not be used in advertising or publicity pertaining to
distribution of the software without specific, written prior permission.

HEWLETT-PACKARD DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE,
INCLUDING ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS, IN NO
EVENT SHALL HEWLETT-PACKARD BE LIABLE FOR ANY SPECIAL, INDIRECT OR
CONSEQUENTIAL DAMAGES OR ANY DAMAGES WHATSOEVER RESULTING FROM LOSS OF
USE, DATA OR PROFITS, WHETHER IN AN ACTION OF CONTRACT, NEGLIGENCE OR
OTHER TORTIOUS ACTION, ARISING OUT OF OR IN CONNECTION WITH THE USE OR
PERFORMANCE OF THIS SOFTWARE.

******************************************************************/
#if !NO_VERSION_STRING
static char version[] = "@(#) $Header: selectinfo.c,v 2.1 93/09/27 11:13:17 chan Exp $";
#endif /* NO_VERSION_STRING */

#include <stdlib.h>
#include <stdio.h>
#include <time.h>
#include <strings.h>

#include <X11/Xlib.h>
#include <X11/Xutil.h>
#include <X11/Intrinsic.h>
#include <X11/Xatom.h>
#include <Xm/Xm.h>
#include <Xm/Protocols.h>
#include <Xm/MainW.h>
#include <Xm/Form.h>
#include <Xm/Label.h>
#include <Xm/List.h>
#include <Xm/PushB.h>
#include <Xm/MessageB.h>


#define MAX_ARGS	25

/* 
   Callback Declarations
*/

void updateCB ();
void targetCB ();
void timeCB ();
void quitCB ();

/*
    Other Routines
*/

void noSelection();
void intern_atoms();
void error_dialog();
int  ignoreXError();
char  *name_from_win();


/*
   Convenient Globals
*/

char		*progname;
Display		*dpy;
int		screen;
XtAppContext     app_context;

Atom      xa_WM_DELETE_WINDOW;
Atom	  xa_TARGETS;
Atom	  xa_TIMESTAMP;
Atom      selection_atom;

char      *owner_name;
Window    owner = NULL;
Time      sel_time = CurrentTime;

Widget 	toplevel,
	name_label, 
	window_label, 
	time_label, 
	type_label, 
	target_label,
	target_list,
	update_btn;


void main (argc, argv)
unsigned int argc; 
char **argv; 
{
    Widget      main_win;
    Widget      cmd_win;
    Widget      value_win;
    char        **original_argv;
    int         original_argc;
    Arg         arg[MAX_ARGS];
    int         n;
    int         i;
    XmString    tmpxmstr;
    
    /*
       We have to save the original argv and argc so that when
       we create our toplevel shell the WM_COMMAND property will
       get set correctly. XtOpenDisplay modifies them so we need
       to do the copying before we call that function.
    */
    original_argv = (char **) XtMalloc((unsigned) (argc + 1) * sizeof(char *));
    for (i = 0; i < argc; i++)
	original_argv[i] = argv[i];
    original_argv[i] = NULL;
    original_argc = argc;

    progname = argv[0];

    /*
       Initialize the toolkit, open the display, and create a 
       toplevel window.
    */
    XtToolkitInitialize();
    app_context = XtCreateApplicationContext();
    dpy = XtOpenDisplay (app_context, NULL, argv[0], "SelectInfo", 
			 NULL, 0, &argc, argv);
    if (!dpy) {
	XtWarning ("appname: Can't open display, exiting...");
	exit (0);
    }
    screen = DefaultScreen(dpy);

    /*
   	Decide which selection to deal with based on name used to 
	invoke program.
    */
    if (strstr(progname, "clip"))
    	intern_atoms("CLIPBOARD");
    else
    	intern_atoms("PRIMARY");

    n = 0;
    XtSetArg(arg[n], XmNallowShellResize, True); n++;
    XtSetArg(arg[n], XmNdeleteResponse, XmDO_NOTHING); n++;
    /* Need to save set argv, argc */
    toplevel = XtAppCreateShell (argv[0], "SelectInfo",
				applicationShellWidgetClass, dpy, arg, n);

    /*
       Set it up so we find out when the window's closed
    */
    xa_WM_DELETE_WINDOW = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    XmAddWMProtocolCallback(toplevel, xa_WM_DELETE_WINDOW, quitCB, NULL);


    /*
	Create our wonderful, massive widget hierarchy (depth first).
    */

    n = 0;
    main_win = XmCreateMainWindow(toplevel, "main", arg, n);
    XtManageChild(main_win);

    n = 0;
    cmd_win = XmCreateForm(main_win, "cmd", arg, n);
    XtManageChild(cmd_win);

    n = 0;
    tmpxmstr = XmStringCreateSimple("Owner Name:             ");
    XtSetArg(arg[n], XmNlabelString, tmpxmstr); n++;
    name_label = XmCreateLabel(cmd_win, "name", arg, n);
    XtManageChild(name_label);
    XmStringFree(tmpxmstr);

    n = 0;
    tmpxmstr = XmStringCreateSimple("Owner Window:           ");
    XtSetArg(arg[n], XmNlabelString, tmpxmstr); n++;
    window_label = XmCreateLabel(cmd_win, "win", arg, n);
    XtManageChild(window_label);
    XmStringFree(tmpxmstr);

    n = 0;
    tmpxmstr = XmStringCreateSimple("Timestamp:              ");
    XtSetArg(arg[n], XmNlabelString, tmpxmstr); n++;
    time_label = XmCreateLabel(cmd_win, "time", arg, n);
    XtManageChild(time_label);
    XmStringFree(tmpxmstr);

    n = 0;
    tmpxmstr = XmStringCreateSimple("Available Targets:      ");
    XtSetArg(arg[n], XmNlabelString, tmpxmstr); n++;
    target_label = XmCreateLabel(cmd_win, "target_label", arg, n);
    XtManageChild(target_label);
    XmStringFree(tmpxmstr);

    n = 0;
    target_list = XmCreateScrolledList(cmd_win, "target_list", arg, n);
    XtManageChild(target_list);

    n = 0;
    update_btn = XmCreatePushButton(cmd_win, "update", arg, n);
    XtManageChild(update_btn);
    XtAddCallback(update_btn, XmNactivateCallback, updateCB, NULL);

    XmMainWindowSetAreas(main_win, NULL, NULL, NULL, NULL, cmd_win);

    XtRealizeWidget (toplevel);
    
#ifdef DEBUG
    XtAppSetSelectionTimeout(app_context, 200000);
#endif

    XtAppMainLoop (app_context);
}


static char      tmpstr[200];

/*
 *   The user has asked that we update our display of selection
 *   information.
 */

void updateCB (w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data;
{
    XmPushButtonCallbackStruct  *cbs = (XmPushButtonCallbackStruct *)call_data;
    Arg       arg[1];
    XmString  tmpxmstr;
    Time      update_time;

    update_time = cbs->event->xbutton.time;

    /*
       First, get the owning window and its name.
    */

    owner = XGetSelectionOwner(dpy, selection_atom);
    if (!owner) {
	noSelection();
	return; /* What should I do???? */
    }
    sprintf(tmpstr, "Owner Window: 0x%x ", owner);
    tmpxmstr = XmStringCreateSimple(tmpstr);
    XtSetArg(arg[0], XmNlabelString, tmpxmstr);
    XtSetValues(window_label, arg, 1);
    XmStringFree(tmpxmstr);

    owner_name = name_from_win(owner);
    /*  Should check length here */
    sprintf(tmpstr, "Owner Name: %s ", owner_name);
    tmpxmstr = XmStringCreateSimple(tmpstr);
    XtSetArg(arg[0], XmNlabelString, tmpxmstr);
    XtSetValues(name_label, arg, 1);
    XmStringFree(tmpxmstr);
    free(owner_name);

    /*
       Next, request TIMESTAMP and TARGETS info.
    */
    XtGetSelectionValue(toplevel, selection_atom, xa_TIMESTAMP, timeCB, 
			NULL, update_time);

    XtGetSelectionValue(toplevel, selection_atom, xa_TARGETS, targetCB,
			NULL, update_time);

}



/* 
   Callback for when we get the TIMESTAMP for the current selection.
*/

void timeCB (w, client_data, selection, type, data, length, format)
Widget w;
XtPointer client_data;
Atom *selection;
Atom *type;
XtPointer data;
unsigned long  *length;
int *format;
{
    Arg       arg[1];
    XmString  tmpxmstr;

#ifdef OLDDEBUG
    fprintf(stderr, "In timeCB, length of selection is %d value is 0x%x\n", *length, *(Time *)data);
#endif

    if (*type == None) {
	/* Bad client doesn't support TIMESTAMP target */
	error_dialog("Selection Owner Doesn't Support TIMESTAMP target", False);
        sprintf(tmpstr, "Timestamp: Not Available");
    } else {
	sprintf(tmpstr, "Timestamp: 0x%x ", *(Time *)data);
    }
    tmpxmstr = XmStringCreateSimple(tmpstr);
    XtSetArg(arg[0], XmNlabelString, tmpxmstr);
    XtSetValues(time_label, arg, 1);
    XmStringFree(tmpxmstr);

}


/*
   Callback for when we get the TARGETS for the current selection.
*/
void targetCB (w, client_data, selection, type, data, length, format)
Widget w;
XtPointer client_data;
Atom *selection;
Atom *type;
XtPointer data;
unsigned long  *length;
int *format;
{
    int       i;
    Arg       arg[1];
    XmString  tmpxmstr;
    char      *atom_name;
    Atom      *atoms = (Atom *)data;

#ifdef OLDDEBUG
    fprintf(stderr, "In targetCB, length of selection is %d\n", *length);
#endif

    XmListDeleteAllItems(target_list);

    if (*type == None) {
	/* Bad client doesn't support TARGETS target */
	error_dialog("Selection Owner Doesn't Support TARGETS target", False);
	return;
    }

    /* Protect against programs giving us garbage in the targets list */
    XSetErrorHandler(ignoreXError);
    for (i = 0; i < *length; i++) {
	atom_name = XGetAtomName(dpy, atoms[i]);
	tmpxmstr = XmStringCreateSimple(atom_name);
	XmListAddItemUnselected(target_list, tmpxmstr, 0);
	XmStringFree(tmpxmstr);
	XFree(atom_name);
    }
    XSetErrorHandler(0);
}




/*
	quitCB			- callback for quit button
*/
void quitCB (w, client_data, call_data)
Widget w;
XtPointer client_data;
XtPointer call_data; 
{

    /*	
	Terminate the application.
	Need to do more ???
     */
    exit (0);
}


/*
   Called when there is no selection owner. Clears out fields of
   display widgets.
*/
void noSelection()
{
    XmString   tmpxmstr;
    Arg        arg[1];

    sprintf(tmpstr, "Owner Window: None ");
    tmpxmstr = XmStringCreateSimple(tmpstr);
    XtSetArg(arg[0], XmNlabelString, tmpxmstr);
    XtSetValues(window_label, arg, 1);
    XmStringFree(tmpxmstr);

    sprintf(tmpstr, "Owner Name: None  ");
    tmpxmstr = XmStringCreateSimple(tmpstr);
    XtSetArg(arg[0], XmNlabelString, tmpxmstr);
    XtSetValues(name_label, arg, 1);
    XmStringFree(tmpxmstr);

    sprintf(tmpstr, "Timestamp:      ");
    tmpxmstr = XmStringCreateSimple(tmpstr);
    XtSetArg(arg[0], XmNlabelString, tmpxmstr);
    XtSetValues(time_label, arg, 1);
    XmStringFree(tmpxmstr);

    XmListDeleteAllItems(target_list);
}



/************************************************************

			Utilities

************************************************************/


/* 
 *  Intern all the atoms we'll need. The one argument is a string 
 *  naming the selection we'll be tracking.
 */

void intern_atoms(selection_name)
char *selection_name;
{
    xa_WM_DELETE_WINDOW = XInternAtom(dpy, "WM_DELETE_WINDOW", False);
    xa_TARGETS = XInternAtom(dpy, "TARGETS", False);
    xa_TIMESTAMP = XInternAtom(dpy, "TIMESTAMP", False);
    selection_atom = XInternAtom(dpy, selection_name, False);
}



/*
 *   Retrieve a name for a given window by tracing up
 *   the window hierarchy until we find a window with
 *   a wm_name hint.
 *
 *   Callers are responsible for freeing the returned string
 *   when it is no longer needed.
 */

#define DEFAULT_OWNER_NAME	"Unknown"

char 
*name_from_win (win)
Window win;
{
    char   *name;
    Window  root, parent, *kids;
    XClassHint   class_hint;
    unsigned int     n;

    /* 
       Now, trace down the window tree to find the window that's
       got some identifying info.
    */
    root = NULL;
    while (win != root) {
	if (XGetClassHint(dpy, win, &class_hint)) {
	    /* Okay, we got our name, return it */
	    XFree(class_hint.res_class);
	    return (class_hint.res_name);
	}
	XQueryTree(dpy, win, &root, &parent, &kids, &n);
	if (n)
	    XFree((char *)kids);
	win = parent;
    }

    /* 
	If we get here, then we never found a window with a 
	class hint, but we want to distinguish from the case
	where the window has gone away
    */
    name = (char *)malloc(strlen(DEFAULT_OWNER_NAME) + 1);
    strcpy(name, DEFAULT_OWNER_NAME);
    return (name);
}
  

int ignoreXError(dpy, err)
Display *dpy;
XErrorEvent *err;
{
    return(0);
}



/*
    Put up a dialog with the given message.
*/

void error_dialog (str, fatal)
char *str;
Boolean fatal;
{
    Widget    dialog;
    Arg	      arg[5];
    XmString  tmpxmstr;
    int	      n;

    n = 0;
    tmpxmstr = XmStringCreateSimple(str);
    XtSetArg(arg[n], XmNmessageString, tmpxmstr); n++;
    XtSetArg(arg[n], XmNdeleteResponse, XmDESTROY); n++;
    dialog = XmCreateMessageDialog(toplevel, "error_dialog", arg, n);
    XmStringFree(tmpxmstr);

    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_CANCEL_BUTTON));
    XtUnmanageChild(XmMessageBoxGetChild(dialog, XmDIALOG_HELP_BUTTON));
    XtManageChild(dialog);

}

